//---------------------------------------------------------------------------
// xspline.cc
//
// this file contains a class XSpline that implements the general XSplines
// presented by Blanc and Schlick in the article 
//
// ``A Spline Model Designed for the End-User.''
//
// This program requires GLUT and OpenGl
//
// Comments and bug-reports to jab@gk.dtu.dk
//---------------------------------------------------------------------------

#include <math.h>
#include <iostream.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "vector.h"

// Hopefully the function below is faster than the 
// generic pow function for integer powers.
inline double intpow(const double x, const int n )
{
  double val = x;
  for (int i=1;i<n;i++)
    val *= x;
  return val;
}


// g(u) as defined in the above mentioned article
double g(double u, double q, double p) 
{
  return 
    q *                                    u 
      + 2 * q *                     intpow(u,2)
	+ (10 - 12 * q - p) *       intpow(u,3)
	  + (2 * p + 14 * q - 15) * intpow(u,4)
	    + (6 - 5*q - p) *       intpow(u,5);
}


// h(u)
double h(double u, double q)
{
  return 
    q *                    u
      + 2 * q *     intpow(u,2)  
	- 2 * q *   intpow(u,4)
	  - q *     intpow(u,5);
}


// XSpline exception enum
enum XSplineErr {XS_BAD_PARAM};


//---------------------------------------------------------------------------
// The XSpline class
//---------------------------------------------------------------------------

class XSpline
{
  int knots;        // Number of knots 
  double DELTA;     // Distance between each

  double inc;       // change sk in inc increments
  double *sv;       // Weight vector
  Vector *cv;       // Control vector

public:

  XSpline();
  ~XSpline();
   
  Vector eval(double t);
  int no_knots() const {return knots;}
  const Vector& get_cp(int k) const {return cv[k];}
  void set_cp(int k, const Vector& pos) {cv[k]=pos;}
  void inc_sk(int k) {if (sv[k]<1.0-inc) sv[k] += inc;}
  void dec_sk(int k) {if (sv[k]>-1.0+inc) sv[k] -= inc;}
};


XSpline::XSpline()
{
  DELTA = 1.0;
  knots = 10;
  inc = 0.01;
  double _sv[] = {0,1,-1,1,-1, 1,1,1,1,0};
  Vector _cv[] =  
  {
    Vector(1,2,0), Vector(2,2,0),  Vector(2,7,0), Vector(4,9,0),
    Vector(6,7,0),  Vector(6,2,0), Vector(7,2,0), Vector(7,1,0),
    Vector(1,1,0),  Vector(1,2,0),
  };
  sv = new double[knots];
  cv = new Vector[knots];
  memcpy(sv,_sv,sizeof(double)*knots);
  memcpy(cv,_cv,sizeof(Vector)*knots);
}
  
XSpline::~XSpline() 
{
  delete [] sv;
  delete [] cv;
}

Vector XSpline::eval(double t) 
{
  if (t>0 && t < knots-1.0) // If we are within the parameter interval
    {
      // First we should find our parameter interval
      // 
      // In accordance with the article, this function defines the value
      // of the spline in the range [ tk+1, tk+2 ], so k is ...
      double kf = floor(t) - 1.0;
      int k = int(kf);

      // Find the big T values
      // 
      // T0p is the point where fk decreases to 0
      // T1p is the point where fk+1 decreases to 0
      // T2m is the point where fk+2 increases from 0
      // T3m is the point where fk+3 increases from 0
      //
      // The parameter sk of a control point k affects
      // the functions that have max vals in k-1 and k+1
      // Therefore, sv[k+1] affects T0p (The point where
      // fk decreases to 0) and T2m. Similarly, sv[k+2]
      // affects T1p and T3m
      double T0p = kf+1 + (sv[k+1]>0?sv[k+1]:0) * DELTA;
      double T1p = kf+2 + (sv[k+2]>0?sv[k+2]:0) * DELTA;
      double T2m = kf+1 - (sv[k+1]>0?sv[k+1]:0) * DELTA;
      double T3m = kf+2 - (sv[k+2]>0?sv[k+2]:0) * DELTA;
	  
      // Find the pk values
      // The pm values are derived directly from the Ts above
      double tmp_factor = 2/(DELTA*DELTA);
      double pm1 = intpow(kf   - T0p, 2) * tmp_factor; 
      double pm0 = intpow(kf+1 - T1p, 2) * tmp_factor; 
      double pp1 = intpow(kf+2 - T2m, 2) * tmp_factor; 
      double pp2 = intpow(kf+3 - T3m, 2) * tmp_factor; 


      // The sv indices are determined like for the Ts
      double qp0 = sv[k+1]<0?-sv[k+1]/2.0:0;
      double qp1 = sv[k+2]<0?-sv[k+2]/2.0:0;
      double qp2 = sv[k+1]<0?-sv[k+1]/2.0:0;
      double qp3 = sv[k+2]<0?-sv[k+2]/2.0:0;
	

      // The code below resembles the part of (17) in the article, 
      // where A0 ... A3 are calculated, but there are differences
      // 
      // 1.
      // 
      // We use g instead of f. This is ok, because 
      //    q = 0 => g(u) = f(u) 
      //
      // 2.
      // we don't check whether we are inside the legal range 
      // of fk+1 and fk+2. This is not necessary, since we are in the
      // range [ tk+1, tk+2 ] and both gk+1 and gk+2 must be defined 
      // in this range.
      // 
      // 3.
      //
      // if we are in the ranges where fk and fk+3 would have been
      // nil in the case of basic splines, we use h instead of g
      // except if q
      //
      double A0,A1,A2,A3;      
      if (t <= T0p)
	A0 =  g((t-T0p)/(kf   - T0p), qp0, pm1);
      else
	A0 = qp0>0 ? h((t-T0p)/(kf   - T0p), qp0) : 0;

      A1 = g((t-T1p)/(kf+1 - T1p), qp1, pm0);

      A2 = g((t-T2m)/(kf+2 - T2m), qp2, pp1);

      if (t >= T3m)
	A3 = g((t-T3m)/(kf+3 - T3m), qp3, pp2);
      else
	A3 = qp3>0 ? h((t-T3m)/(kf+3 - T3m), qp3) : 0;

      double SUM = A0+A1+A2+A3;

      Vector C =  cv[k]*A0+cv[k+1]*A1+cv[k+2]*A2+cv[k+3]*A3;
      C /= SUM; // Normalise
      return C;
    }
  // We should throw something ... a rotten apple or exception.
  throw XS_BAD_PARAM;
}


//---------------------------------------------------------------------------
// GLOBALS
//---------------------------------------------------------------------------

XSpline C;        // our spline
int last_picked;  // last picked control point
//---------------------------------------------------------------------------


void reset_proj() {  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0,10,0,10); 
  glMatrixMode(GL_MODELVIEW);
}

void display()
{
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  glColor3f(0,0,.3);  

  for (int j=0; j< C.no_knots(); j++)
    {
      Vector pos = C.get_cp(j);
      double d = .1;
      glBegin(GL_POLYGON);
      glVertex3d(pos[0]-d,pos[1]-d,pos[2]);
      glVertex3d(pos[0]+d,pos[1]-d,pos[2]);
      glVertex3d(pos[0]+d,pos[1]+d,pos[2]);
      glVertex3d(pos[0]-d,pos[1]+d,pos[2]);
      glEnd();
    }


  
  const int segs = 300;
  bool colb = false;
  glBegin(GL_LINE_STRIP);
  for (int i=0; i<=segs; i++)
    {
      if(colb) 
	glColor3f(1,0,0);    
      else
	glColor3f(0,1,0);
      colb = !colb;
      double t = 10.0 * double(i)/double(segs);
      try
	{
	  Vector pos = C.eval(t);
	  glVertex3d(pos[0],pos[1],pos[2]);
	}
      catch(XSplineErr) 
	{
	  //cerr << "spline not defined for t = " << t << endl; 
	};
    }
  glEnd();
  
}


static void key( unsigned char key, int x, int y )
{
  switch (key) {
  case 27:
    exit(0);
    break;
  case '+':
    C.inc_sk(last_picked);
    glutPostRedisplay();
    break;
  case '-':
    C.dec_sk(last_picked);
    glutPostRedisplay();
    break;

  }
}

bool LeftButtonPressed = false;
void mouse_button_fun(int button, int state, int mousex, int mousey)
{
  
  GLint viewport[4];
  glGetIntegerv(GL_VIEWPORT, viewport);

  if(state == GLUT_DOWN)
    {
      const int bufsize = 512;
      unsigned select_buffer[bufsize];
      memset(select_buffer,0,512*sizeof(int));
      GLint hits;
      // Switch to select mode, and set up selection buffer
      glSelectBuffer(bufsize, select_buffer);

      glInitNames();

      // Setup projection matrix
      

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPickMatrix(mousex, viewport[3] - mousey, 2,2, viewport);
      gluOrtho2D(0,10,0,10);
      
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

      glRenderMode(GL_SELECT);
      for (int j=0; j< C.no_knots(); j++)
	{
	  Vector pos = C.get_cp(j);
	  double d = 0.1;
	  glPushName(j);
	  glBegin(GL_POLYGON);
	  glVertex2f(pos[0]-d,pos[1]-d);
	  glVertex2f(pos[0]+d,pos[1]-d);
	  glVertex2f(pos[0]+d,pos[1]+d);
	  glVertex2f(pos[0]-d,pos[1]+d);
	  glEnd();
	  glFlush();
	  glPopName();
	}
      hits = glRenderMode(GL_RENDER);

      
      reset_proj();

      if(hits)
	{
	  last_picked = select_buffer[3];
	}      
      LeftButtonPressed = true;
    }
  else
    {
      Vector pos;

      pos[0] = 10 * double(mousex)/double(viewport[2]);
      pos[1] = 10 * double(viewport[3] - mousey)/double(viewport[3]);
      C.set_cp(last_picked, pos);
      glutPostRedisplay();
      
      LeftButtonPressed = false;
    }
}

void mouse_move_fun(int mousex, int mousey)
{
  if (LeftButtonPressed)
    {
      Vector pos;
      GLint viewport[4];
      
      glGetIntegerv(GL_VIEWPORT, viewport);

      pos[0] = 10 * double(mousex)/double(viewport[2]);
      pos[1] = 10 * double(viewport[3] - mousey)/double(viewport[3]);
      C.set_cp(last_picked, pos);
      glutPostRedisplay();
    }
}

static void reshape( int width, int height )
{
  glViewport( 0, 0, width, height );
}


void main(int argc, char *argv[])
{
  glutInit( &argc, argv );
  glutInitWindowSize( 500 , 500 );
  glutCreateWindow(argv[0]);
  glLineWidth(2.0);
  glClearColor(.7,.7,.7,1);
  glShadeModel(GL_FLAT);
  reset_proj();
  glutReshapeFunc( reshape );
  glutKeyboardFunc( key );
  glutDisplayFunc( display );
  glutMouseFunc( mouse_button_fun );
  glutMotionFunc( mouse_move_fun );

  glutMainLoop();
  
}





